@@ -3,7 +3,7 @@
  * Copyright (c) 1989-1992, Brian Berliner
  * 
  * You may distribute under the terms of the GNU General Public License as
- * specified in the README file that comes with the CVS 1.3 kit.
+ * specified in the README file that comes with the CVS 1.4 kit.
  * 
  * Create Version
  * 
@@ -36,24 +36,19 @@
 #include "cvs.h"
 
 #ifndef lint
-static char rcsid[] = "@(#)checkout.c 1.67 92/04/10";
+static const char rcsid[] = "$CVSid: @(#)checkout.c 1.78 94/10/07 $";
+USE(rcsid)
 #endif
 
-#if __STDC__
-static char *findslash (char *start, char *p);
-static int build_dirs_and_chdir (char *dir, char *prepath, char *realdir,
-				 int sticky);
-static int checkout_proc (int *pargc, char *argv[], char *where,
+static char *findslash PROTO((char *start, char *p));
+static int build_dirs_and_chdir PROTO((char *dir, char *prepath, char *realdir,
+				 int sticky));
+static int checkout_proc PROTO((int *pargc, char **argv, char *where,
 		          char *mwhere, char *mfile, int shorten,
 		          int local_specified, char *omodule,
-		          char *msg);
-#else
-static int checkout_proc ();
-static char *findslash ();
-static int build_dirs_and_chdir ();
-#endif				/* __STDC__ */
+		          char *msg));
 
-static char *checkout_usage[] =
+static const char *const checkout_usage[] =
 {
     "Usage:\n  %s %s [-ANPQcflnpqs] [-r rev | -D date] [-d dir] [-k kopt] modules...\n",
     "\t-A\tReset any sticky tags/date/kopts.\n",
@@ -75,7 +70,7 @@ static char *checkout_usage[] =
     NULL
 };
 
-static char *export_usage[] =
+static const char *const export_usage[] =
 {
     "Usage: %s %s [-NPQflnq] [-r rev | -D date] [-d dir] module...\n",
     "\t-N\tDon't shorten module paths if -d specified.\n",
@@ -84,8 +79,8 @@ static char *export_usage[] =
     "\t-l\tLocal directory only, not recursive\n",
     "\t-n\tDo not run module program (if any).\n",
     "\t-q\tSomewhat quiet.\n",
-    "\t-r rev\tCheck out revision or tag. (implies -P)\n",
-    "\t-D date\tCheck out revisions as of date. (implies -P)\n",
+    "\t-r rev\tCheck out revision or tag.\n",
+    "\t-D date\tCheck out revisions as of date.\n",
     "\t-d dir\tCheck out into dir instead of module name.\n",
     NULL
 };
@@ -104,9 +99,9 @@ static char *preload_update_dir = NULL;
 int
 checkout (argc, argv)
     int argc;
-    char *argv[];
+    char **argv;
 {
-    register int i;
+    int i;
     int c;
     DBM *db;
     int cat = 0, err = 0, status = 0;
@@ -114,7 +109,8 @@ checkout (argc, argv)
     int local = 0;
     int shorten = -1;
     char *where = NULL;
-    char *valid_options, **valid_usage;
+    char *valid_options;
+    const char *const *valid_usage;
 
     /*
      * A smaller subset of options are allowed for the export command, which
@@ -139,7 +135,7 @@ checkout (argc, argv)
     ign_setup ();
 
     optind = 1;
-    while ((c = gnu_getopt (argc, argv, valid_options)) != -1)
+    while ((c = getopt (argc, argv, valid_options)) != -1)
     {
 	switch (c)
 	{
@@ -238,6 +234,88 @@ checkout (argc, argv)
 	options = RCS_check_kflag ("v");/* -kv must be on */
     }
 
+    if (client_active)
+    {
+	int expand_modules;
+
+	start_server ();
+
+	ign_setup ();
+	
+	/* We have to expand names here because the "expand-modules"
+           directive to the server has the side-effect of having the
+           server send the check-in and update programs for the
+           various modules/dirs requested.  If we turn this off and
+           simply request the names of the modules and directories (as
+           below in !expand_modules), those files (CVS/Checking.prog
+           or CVS/Update.prog) don't get created.  Grrr.  */
+	
+	expand_modules = (!cat && !status && !pipeout
+			  && supported_request ("expand-modules"));
+	
+	if (expand_modules)
+	  {
+	    /* This is done here because we need to read responses
+               from the server before we send the command checkout or
+               export files. */
+
+	    client_expand_modules (argc, argv, local);
+	  }
+
+	if (!run_module_prog) send_arg ("-n");
+	if (really_quiet) send_arg ("-Q");
+	if (quiet) send_arg ("-q");
+	if (local) send_arg ("-l");
+	if (pipeout) send_arg ("-p");
+	if (!force_tag_match) send_arg ("-f");
+	if (aflag)
+	    send_arg("-A");
+	if (!shorten)
+	    send_arg("-N");
+	if (checkout_prune_dirs && strcmp (command_name, "export") != 0)
+	    send_arg("-P");
+	client_prune_dirs = checkout_prune_dirs;
+	if (cat)
+	    send_arg("-c");
+	if (where != NULL)
+	{
+	    option_with_arg ("-d", where);
+	}
+	if (status)
+	    send_arg("-s");
+	if (strcmp (command_name, "export") != 0
+	    && options != NULL
+	    && options[0] != '\0')
+	    send_arg (options);
+	option_with_arg ("-r", tag);
+	if (date)
+	    client_senddate (date);
+	if (join_rev1 != NULL)
+	    option_with_arg ("-j", join_rev1);
+	if (join_rev2 != NULL)
+	    option_with_arg ("-j", join_rev2);
+
+	if (expand_modules)
+	  {
+	    client_send_expansions (local);
+	  }
+	else
+	  {
+	    int i;
+	    for (i = 0; i < argc; ++i)
+	      send_arg (argv[i]);
+	    client_nonexpanded_setup ();
+	  }
+
+	if (fprintf
+	    (to_server,
+	     strcmp (command_name, "export") == 0 ? "export\n" : "co\n")
+	    < 0)
+	  error (1, errno, "writing to server");
+
+	return get_responses_and_close ();
+    }
+
     if (cat || status)
     {
 	cat_module (status);
@@ -259,12 +337,19 @@ checkout (argc, argv)
 	    error (1, errno, "cannot chdir to %s", where);
 	preload_update_dir = xstrdup (where);
 	where = (char *) NULL;
-	if (!isfile (CVSADM) && !isfile (OCVSADM))
+	if (!isfile (CVSADM))
 	{
-	    (void) sprintf (repository, "%s/%s", CVSroot, CVSNULLREPOS);
+	    (void) sprintf (repository, "%s/%s/%s", CVSroot, CVSROOTADM,
+			    CVSNULLREPOS);
 	    if (!isfile (repository))
 		(void) mkdir (repository, 0777);
-	    Create_Admin (".", repository, (char *) NULL, (char *) NULL);
+
+	    /* I'm not sure whether this check is redundant.  */
+	    if (!isdir (repository))
+		error (1, 0, "there is no repository %s", repository);
+
+	    Create_Admin (".", where, repository,
+			  (char *) NULL, (char *) NULL);
 	    if (!noexec)
 	    {
 		FILE *fp;
@@ -272,6 +357,8 @@ checkout (argc, argv)
 		fp = open_file (CVSADM_ENTSTAT, "w+");
 		if (fclose(fp) == EOF)
 		    error(1, errno, "cannot close %s", CVSADM_ENTSTAT);
+		if (server_active)
+		    server_set_entstat (preload_update_dir, repository);
 	    }
 	}
     }
@@ -283,11 +370,11 @@ checkout (argc, argv)
      * attempt to cd to the indicated place.  where then becomes simply the
      * last component
      */
-    if (where != NULL && index (where, '/') != NULL)
+    if (where != NULL && strchr (where, '/') != NULL)
     {
 	char *slash;
 
-	slash = rindex (where, '/');
+	slash = strrchr (where, '/');
 	*slash = '\0';
 
 	if (chdir (where) < 0)
@@ -316,7 +403,7 @@ static int
 checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
 	       local_specified, omodule, msg)
     int *pargc;
-    char *argv[];
+    char **argv;
     char *where;
     char *mwhere;
     char *mfile;
@@ -358,7 +445,7 @@ checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
 	char file[PATH_MAX];
 
 	/* if mfile is really a path, straighten it out first */
-	if ((cp = rindex (mfile, '/')) != NULL)
+	if ((cp = strrchr (mfile, '/')) != NULL)
 	{
 	    *cp = 0;
 	    (void) strcat (repository, "/");
@@ -386,7 +473,7 @@ checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
 	    {
 		char *slash;
 
-		if ((slash = rindex (mfile, '/')) != NULL)
+		if ((slash = strrchr (mfile, '/')) != NULL)
 		    mwhere = slash + 1;
 		else
 		    mwhere = mfile;
@@ -448,7 +535,7 @@ checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
      */
     if (shorten && where == NULL)
     {
-	if ((cp = rindex (argv[0], '/')) != NULL)
+	if ((cp = strrchr (argv[0], '/')) != NULL)
 	{
 	    (void) strcpy (xwhere, cp + 1);
 	    where = xwhere;
@@ -498,8 +585,8 @@ checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
 	 * elements exist in where. Big Black Magic
 	 */
 	prepath = xstrdup (repository);
-	cp = rindex (where, '/');
-	cp2 = rindex (prepath, '/');
+	cp = strrchr (where, '/');
+	cp2 = strrchr (prepath, '/');
 	while (cp != NULL)
 	{
 	    cp = findslash (where, cp - 1);
@@ -527,19 +614,32 @@ checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
 	free (prepath);
 
 	/* set up the repository (or make sure the old one matches) */
-	if (!isfile (CVSADM) && !isfile (OCVSADM))
+	if (!isfile (CVSADM))
 	{
 	    FILE *fp;
 
 	    if (!noexec && *pargc > 1)
 	    {
-		Create_Admin (".", repository, (char *) NULL, (char *) NULL);
+		/* I'm not sure whether this check is redundant.  */
+		if (!isdir (repository))
+		    error (1, 0, "there is no repository %s", repository);
+
+		Create_Admin (".", where, repository,
+			      (char *) NULL, (char *) NULL);
 		fp = open_file (CVSADM_ENTSTAT, "w+");
 		if (fclose(fp) == EOF)
 		    error(1, errno, "cannot close %s", CVSADM_ENTSTAT);
+		if (server_active)
+		    server_set_entstat (where, repository);
 	    }
 	    else
-		Create_Admin (".", repository, tag, date);
+	    {
+		/* I'm not sure whether this check is redundant.  */
+		if (!isdir (repository))
+		    error (1, 0, "there is no repository %s", repository);
+
+		Create_Admin (".", where, repository, tag, date);
+	    }
 	}
 	else
 	{
@@ -609,7 +709,7 @@ checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
 	List *entries;
 
 	/* we are only doing files, so register them */
-	entries = ParseEntries (0);
+	entries = Entries_Open (0);
 	for (i = 1; i < *pargc; i++)
 	{
 	    char line[MAXLINELEN];
@@ -622,12 +722,14 @@ checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
 	    if (vers->ts_user == NULL)
 	    {
 		(void) sprintf (line, "Initial %s", user);
-		Register (entries, user, vers->vn_rcs, line, vers->options,
-			  vers->tag, vers->date);
+		Register (entries, user, vers->vn_rcs ? vers->vn_rcs : "0",
+			  line, vers->options, vers->tag,
+			  vers->date, (char *) 0);
 	    }
 	    freevers_ts (&vers);
 	}
-	dellist (&entries);
+
+	Entries_Close (entries);
     }
 
     /* Don't log "export", just regular "checkouts" */
@@ -650,9 +752,9 @@ findslash (start, p)
     char *start;
     char *p;
 {
-    while ((int) p >= (int) start && *p != '/')
+    while (p >= start && *p != '/')
 	p--;
-    if ((int) p < (int) start)
+    if (p < start)
 	return (NULL);
     else
 	return (p);
@@ -681,7 +783,7 @@ build_dirs_and_chdir (dir, prepath, realdir, sticky)
     (void) strcpy (path, dir);
     (void) strcpy (path2, realdir);
     for (cp = path, cp2 = path2;
-    (slash = index (cp, '/')) != NULL && (slash2 = index (cp2, '/')) != NULL;
+    (slash = strchr (cp, '/')) != NULL && (slash2 = strchr (cp2, '/')) != NULL;
 	 cp = slash + 1, cp2 = slash2 + 1)
     {
 	*slash = '\0';
@@ -692,17 +794,21 @@ build_dirs_and_chdir (dir, prepath, realdir, sticky)
 	    error (0, errno, "cannot chdir to %s", cp);
 	    return (1);
 	}
-	if (!isfile (CVSADM) && !isfile (OCVSADM) &&
-	    strcmp (command_name, "export") != 0)
+	if (!isfile (CVSADM) && strcmp (command_name, "export") != 0)
 	{
 	    (void) sprintf (repository, "%s/%s", prepath, path2);
-	    Create_Admin (".", repository, sticky ? (char *) NULL : tag,
+	    /* I'm not sure whether this check is redundant.  */
+	    if (!isdir (repository))
+		error (1, 0, "there is no repository %s", repository);
+	    Create_Admin (".", cp, repository, sticky ? (char *) NULL : tag,
 			  sticky ? (char *) NULL : date);
 	    if (!noexec)
 	    {
 		fp = open_file (CVSADM_ENTSTAT, "w+");
 		if (fclose(fp) == EOF)
 		    error(1, errno, "cannot close %s", CVSADM_ENTSTAT);
+		if (server_active)
+		    server_set_entstat (path, repository);
 	    }
 	}
 	*slash = '/';
