@@ -2,45 +2,63 @@
  * Copyright (c) 1992, Brian Berliner and Jeff Polk
  * 
  * You may distribute under the terms of the GNU General Public License as
- * specified in the README file that comes with the CVS 1.3 kit.
+ * specified in the README file that comes with the CVS 1.4 kit.
  * 
  * The routines contained in this file do all the rcs file parsing and
  * manipulation
  */
 
+#include <assert.h>
 #include "cvs.h"
 
 #ifndef lint
-static char rcsid[] = "@(#)rcs.c 1.28 92/03/31";
+static const char rcsid[] = "$CVSid: @(#)rcs.c 1.40 94/10/07 $";
+USE(rcsid)
 #endif
 
-#if __STDC__
-static char *RCS_getbranch (RCSNode * rcs, char *tag, int force_tag_match);
-static char *RCS_getdatebranch (RCSNode * rcs, char *date, char *branch);
-static int getrcskey (FILE * fp, char **keyp, char **valp);
-static int parse_rcs_proc (Node * file);
-static int checkmagic_proc (Node *p);
-static void do_branches (List * list, char *val);
-static void do_symbols (List * list, char *val);
-static void null_delproc (Node * p);
-static void rcsnode_delproc (Node * p);
-static void rcsvers_delproc (Node * p);
-#else
-static int parse_rcs_proc ();
-static int checkmagic_proc ();
-static void rcsnode_delproc ();
-static void rcsvers_delproc ();
-static void null_delproc ();
-static int getrcskey ();
-static void do_symbols ();
-static void do_branches ();
-static char *RCS_getbranch ();
-static char *RCS_getdatebranch ();
-#endif				/* __STDC__ */
+static RCSNode *RCS_parsercsfile_i PROTO((FILE * fp, const char *rcsfile));
+static char *RCS_getdatebranch PROTO((RCSNode * rcs, char *date, char *branch));
+static int getrcskey PROTO((FILE * fp, char **keyp, char **valp));
+static int parse_rcs_proc PROTO((Node * file, void *closure));
+static int checkmagic_proc PROTO((Node *p, void *closure));
+static void do_branches PROTO((List * list, char *val));
+static void do_symbols PROTO((List * list, char *val));
+static void null_delproc PROTO((Node * p));
+static void rcsnode_delproc PROTO((Node * p));
+static void rcsvers_delproc PROTO((Node * p));
 
 static List *rcslist;
 static char *repository;
 
+/*
+ * We don't want to use isspace() from the C library because:
+ *
+ * 1. The definition of "whitespace" in RCS files includes ASCII
+ *    backspace, but the C locale doesn't.
+ * 2. isspace is an very expensive function call in some implementations
+ *    due to the addition of wide character support.
+ */
+static const char spacetab[] = {
+        0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0,	/* 0x00 - 0x0f */
+        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 0x10 - 0x1f */
+        1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 0x20 - 0x2f */
+        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 0x30 - 0x3f */
+        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 0x40 - 0x4f */
+        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 0x50 - 0x5f */
+        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 0x60 - 0x8f */
+        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 0x70 - 0x7f */
+        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 0x80 - 0x8f */
+        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 0x90 - 0x9f */
+        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 0xa0 - 0xaf */
+        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 0xb0 - 0xbf */
+        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 0xc0 - 0xcf */
+        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 0xd0 - 0xdf */
+        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 0xe0 - 0xef */
+        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0  /* 0xf0 - 0xff */
+};
+
+#define whitespace(c)	(spacetab[c] != 0)
+
 /*
  * Parse all the rcs files specified and return a list
  */
@@ -54,7 +72,7 @@ RCS_parsefiles (files, xrepos)
     rcslist = getlist ();
 
     /* walk the list parsing files */
-    if (walklist (files, parse_rcs_proc) != 0)
+    if (walklist (files, parse_rcs_proc, NULL) != 0)
     {
 	/* free the list and return NULL on error */
 	dellist (&rcslist);
@@ -69,10 +87,10 @@ RCS_parsefiles (files, xrepos)
  * Parse an rcs file into a node on the rcs list
  */
 static int
-parse_rcs_proc (file)
+parse_rcs_proc (file, closure)
     Node *file;
+    void *closure;
 {
-    Node *p;
     RCSNode *rdata;
 
     /* parse the rcs file into rdata */
@@ -80,63 +98,91 @@ parse_rcs_proc (file)
 
     /* if we got a valid RCSNode back, put it on the list */
     if (rdata != (RCSNode *) NULL)
-    {
-	p = getnode ();
-	p->key = xstrdup (file->key);
-	p->delproc = rcsnode_delproc;
-	p->type = RCSNODE;
-	p->data = (char *) rdata;
-	(void) addnode (rcslist, p);
-    }
+	RCS_addnode (file->key, rdata, rcslist);
+
     return (0);
 }
 
+/*
+ * Add an RCSNode to a list of them.
+ */
+
+void
+RCS_addnode (file, rcs, list)
+    const char *file;
+    RCSNode *rcs;
+    List *list;
+{
+    Node *p;
+    
+    p = getnode ();
+    p->key = xstrdup (file);
+    p->delproc = rcsnode_delproc;
+    p->type = RCSNODE;
+    p->data = (char *) rcs;
+    (void) addnode (list, p);
+}
+
+
 /*
  * Parse an rcsfile given a user file name and a repository
  */
 RCSNode *
 RCS_parse (file, repos)
-    char *file;
-    char *repos;
+    const char *file;
+    const char *repos;
 {
     RCSNode *rcs;
+    FILE *fp;
     char rcsfile[PATH_MAX];
 
     (void) sprintf (rcsfile, "%s/%s%s", repos, file, RCSEXT);
-    if (!isreadable (rcsfile))
+    if ((fp = fopen (rcsfile, "r")) != NULL) 
     {
-	(void) sprintf (rcsfile, "%s/%s/%s%s", repos, CVSATTIC,
-			file, RCSEXT);
-	if (!isreadable (rcsfile))
-	    return (NULL);
-	rcs = RCS_parsercsfile (rcsfile);
+        rcs = RCS_parsercsfile_i(fp, rcsfile);
+	if (rcs != NULL) 
+	    rcs->flags |= VALID;
+
+	fclose (fp);
+	return (rcs);
+    }
+    else if (errno != ENOENT)
+    {
+	error (0, errno, "cannot open %s", rcsfile);
+	return NULL;
+    }
+
+    (void) sprintf (rcsfile, "%s/%s/%s%s", repos, CVSATTIC, file, RCSEXT);
+    if ((fp = fopen (rcsfile, "r")) != NULL) 
+    {
+        rcs = RCS_parsercsfile_i(fp, rcsfile);
 	if (rcs != NULL)
 	{
 	    rcs->flags |= INATTIC;
 	    rcs->flags |= VALID;
 	}
+
+	fclose (fp);
 	return (rcs);
     }
-    rcs = RCS_parsercsfile (rcsfile);
-    if (rcs != NULL)
-	rcs->flags |= VALID;
-    return (rcs);
+    else if (errno != ENOENT)
+    {
+	error (0, errno, "cannot open %s", rcsfile);
+	return NULL;
+    }
+
+    return (NULL);
 }
 
 /*
- * Do the real work of parsing an RCS file
+ * Parse a specific rcsfile.
  */
 RCSNode *
 RCS_parsercsfile (rcsfile)
     char *rcsfile;
 {
-    Node *q, *r;
-    RCSNode *rdata;
-    RCSVers *vnode;
-    int n;
-    char *cp;
-    char *key, *value;
     FILE *fp;
+    RCSNode *rcs;
 
     /* open the rcsfile */
     if ((fp = fopen (rcsfile, "r")) == NULL)
@@ -145,11 +191,103 @@ RCS_parsercsfile (rcsfile)
 	return (NULL);
     }
 
+    rcs = RCS_parsercsfile_i (fp, rcsfile);
+
+    fclose (fp);
+    return (rcs);
+}
+
+
+/*
+ */ 
+static RCSNode *
+RCS_parsercsfile_i (fp, rcsfile)
+    FILE *fp;
+    const char *rcsfile;
+{
+    RCSNode *rdata;
+    char *key, *value;
+
     /* make a node */
     rdata = (RCSNode *) xmalloc (sizeof (RCSNode));
-    bzero ((char *) rdata, sizeof (RCSNode));
+    memset ((char *) rdata, 0, sizeof (RCSNode));
     rdata->refcount = 1;
     rdata->path = xstrdup (rcsfile);
+
+    /* Process HEAD and BRANCH keywords from the RCS header.  
+     *
+     * Most cvs operatations on the main branch don't need any more
+     * information.  Those that do call XXX to completely parse the
+     * RCS file.  */
+
+    if (getrcskey (fp, &key, &value) == -1 || key == NULL)
+	goto error;
+
+    if (strcmp (RCSHEAD, key) == 0 && value != NULL)
+	rdata->head = xstrdup (value);
+
+    if (getrcskey (fp, &key, &value) == -1 || key == NULL)
+	goto error;
+
+    if (strcmp (RCSBRANCH, key) == 0 && value != NULL)
+    {
+	char *cp;
+
+	rdata->branch = xstrdup (value);
+	if ((numdots (rdata->branch) & 1) != 0)
+	{
+	    /* turn it into a branch if it's a revision */
+	    cp = strrchr (rdata->branch, '.');
+	    *cp = '\0';
+	}
+    }
+
+    rdata->flags |= PARTIAL;
+    return rdata;
+
+error:
+    if (!really_quiet)
+    {
+	if (ferror(fp))
+	{
+	    error (1, 0, "error reading `%s'", rcsfile);
+	}
+	else
+	{
+	    error (0, 0, "`%s' does not appear to be a valid rcs file",
+		   rcsfile);
+	}
+    }
+    freercsnode (&rdata);
+    return (NULL);
+}
+
+
+/*
+ * Do the real work of parsing an RCS file 
+ *
+ * There are no allowances for error here.
+ */
+void
+RCS_reparsercsfile (rdata)
+    RCSNode *rdata;
+{
+    FILE *fp;
+    char *rcsfile;
+
+    Node *q, *r;
+    RCSVers *vnode;
+    int n;
+    char *cp;
+    char *key, *value;
+
+    rcsfile = rdata->path;
+
+    fp = fopen(rcsfile, "r");
+    if (fp == NULL)
+	error (1, 0, "unable to reopen `%s'", rcsfile);
+
+    /* make a node */
     rdata->versions = getlist ();
     rdata->dates = getlist ();
 
@@ -161,41 +299,26 @@ RCS_parsercsfile (rcsfile)
     {
 	/* get the next key/value pair */
 
-	/* if key is NULL here, then the file is missing some headers */
+	/* if key is NULL here, then the file is missing some headers
+	   or we had trouble reading the file. */
 	if (getrcskey (fp, &key, &value) == -1 || key == NULL)
 	{
-	    if (!really_quiet)
-		error (0, 0, "`%s' does not appear to be a valid rcs file",
+	    if (ferror(fp))
+	    {
+		error (1, 0, "error reading `%s'", rcsfile);
+	    }
+	    else
+	    {
+		error (1, 0, "`%s' does not appear to be a valid rcs file",
 		       rcsfile);
-	    freercsnode (&rdata);
-	    (void) fclose (fp);
-	    return (NULL);
+	    }
 	}
 
-	/* process it */
-	if (strcmp (RCSHEAD, key) == 0 && value != NULL)
-	{
-	    rdata->head = xstrdup (value);
-	    continue;
-	}
-	if (strcmp (RCSBRANCH, key) == 0 && value != NULL)
-	{
-	    rdata->branch = xstrdup (value);
-	    if ((numdots (rdata->branch) & 1) != 0)
-	    {
-		/* turn it into a branch if it's a revision */
-		cp = rindex (rdata->branch, '.');
-		*cp = '\0';
-	    }
-	    continue;
-	}
 	if (strcmp (RCSSYMBOLS, key) == 0)
 	{
 	    if (value != NULL)
 	    {
-		/* if there are tags, set up the tag list */
-		rdata->symbols = getlist ();
-		do_symbols (rdata->symbols, value);
+		rdata->symbols_data = xstrdup(value);
 		continue;
 	    }
 	}
@@ -225,7 +348,7 @@ RCS_parsercsfile (rcsfile)
 
 	/* grab the value of the date from value */
 	valp = value + strlen (RCSDATE);/* skip the "date" keyword */
-	while (isspace (*valp))		/* take space off front of value */
+	while (whitespace (*valp))		/* take space off front of value */
 	    valp++;
 	(void) strcpy (date, valp);
 
@@ -237,7 +360,7 @@ RCS_parsercsfile (rcsfile)
 	q->delproc = rcsvers_delproc;
 	r->delproc = null_delproc;
 	q->data = r->data = xmalloc (sizeof (RCSVers));
-	bzero (q->data, sizeof (RCSVers));
+	memset (q->data, 0, sizeof (RCSVers));
 	vnode = (RCSVers *) q->data;
 
 	/* fill in the version before we forget it */
@@ -248,6 +371,17 @@ RCS_parsercsfile (rcsfile)
 
 	/* throw away the state field */
 	(void) getrcskey (fp, &key, &value);
+#ifdef DEATH_SUPPORT
+	/* Accept this regardless of DEATH_STATE, so that we can read
+	   repositories created with different versions of CVS.  */
+	if (strcmp (key, "state") != 0)
+	    error (1, 0, "\
+unable to parse rcs file; `state' not in the expected place");
+	if (strcmp (value, "dead") == 0)
+	{
+	    vnode->dead = 1;
+	}
+#endif
 
 	/* fill in the date field */
 	r->key = vnode->date = xstrdup (date);
@@ -271,13 +405,89 @@ RCS_parsercsfile (rcsfile)
 	 */
 	while ((n = getrcskey (fp, &key, &value)) >= 0)
 	{
+#ifdef DEATH_SUPPORT
+	    /* Enable use of repositories created with a CVS which defines
+	       DEATH_SUPPORT and not DEATH_STATE or CVSDEA.  */
+	    if (strcmp(key, RCSDEAD) == 0)
+	    {
+		vnode->dead = 1;
+		continue;
+	    }
+#endif
 	    /* if we have a revision, break and do it */
 	    for (cp = key; (isdigit (*cp) || *cp == '.') && *cp != '\0'; cp++)
 		 /* do nothing */ ;
 	    if (*cp == '\0' && strncmp (RCSDATE, value, strlen (RCSDATE)) == 0)
 		break;
 	}
+#ifdef CVSDEA
+	/* Accept this only #ifdef CVSDEA, because (a) all that extra file
+	   access presumably slows things down, (b) I don't think anyone
+	   is actually using CVSDEA.  It probably could be flushed
+	   entirely.  */
+	/* Check whether vnode->version is listed in CVSDEA file.  */
+	{
+	    char *p;
+	    FILE *deafile;
+	    int ch;
+	    char *tmp = xmalloc (sizeof (CVSDEA) + strlen (rcsfile) + 40);
+	    char *last_slash = strrchr (rcsfile, '/');
+	    int len;
 
+	    /* Convert /foo/bar/baz/bif,v to /foo/bar/baz/CVS.dea/bif.  */
+
+	    strncpy (tmp, rcsfile, last_slash - rcsfile + 1);
+	    p = tmp + (last_slash - rcsfile) + 1;
+
+	    /* CVSDEA files do *not* move into the attic.  */
+	    if (last_slash - rcsfile > 6 && strncmp (p - 6, "Attic/", 6) == 0)
+		p -= 6;
+
+	    strcpy (p, CVSDEA);
+	    strcat (p, "/");
+	    p += strlen (p);
+	    len = strlen (last_slash + 1) - (sizeof (RCSEXT) - 1);
+	    strncpy (p, last_slash + 1, len);
+	    p[len] = '\0';
+
+	    deafile = fopen (tmp, "r");
+	    if (deafile == NULL)
+	    {
+		if (errno != ENOENT)
+		    error (1, errno, "cannot open %s", tmp);
+	    }
+	    else
+	    {
+		p = vnode->version;
+		while (1)
+		{
+		    ch = getc (deafile);
+		    if (ferror (deafile))
+			error (1, errno, "cannot read %s", tmp);
+		    if (feof (deafile))
+			break;
+		    if (ch == '\012')
+		    {
+			if (p != NULL && *p == '\0')
+			{
+			    vnode->dead = 1;
+			    break;
+			}
+			p = vnode->version;
+		    }
+		    else
+		    {
+			if (p != NULL && *p++ != ch)
+			    p = NULL;
+		    }
+		}
+		if (fclose (deafile) == EOF)
+		    error (1, errno, "cannot close %s", tmp);
+	    }
+	    free (tmp);
+	}
+#endif /* CVSDEA */
+		
 	/* add the nodes to the lists */
 	(void) addnode (rdata->versions, q);
 	(void) addnode (rdata->dates, r);
@@ -290,8 +500,8 @@ RCS_parsercsfile (rcsfile)
 	    break;
     }
 
-    (void) fclose (fp);
-    return (rdata);
+    fclose (fp);
+    rdata->flags &= ~PARTIAL;
 }
 
 /*
@@ -325,6 +535,8 @@ freercsnode (rnodep)
     dellist (&(*rnodep)->dates);
     if ((*rnodep)->symbols != (List *) NULL)
 	dellist (&(*rnodep)->symbols);
+    if ((*rnodep)->symbols_data != (char *) NULL)
+	free ((*rnodep)->symbols_data);
     if ((*rnodep)->head != (char *) NULL)
 	free ((*rnodep)->head);
     if ((*rnodep)->branch != (char *) NULL)
@@ -370,17 +582,17 @@ null_delproc (p)
  *      space or semicolon 
  *    o if key == "desc" then key and data are NULL and return -1 
  *    o if key wasn't terminated by a semicolon, skip white space and fill 
- *      in value with everything up to a semicolon o compress all whitespace
- *      down to a single space 
+ *      in value with everything up to a semicolon 
+ *    o compress all whitespace down to a single space 
  *    o if a word starts with @, do funky rcs processing
  *    o strip whitespace off end of value or set value to NULL if it empty 
  *    o return 0 since we found something besides "desc"
  */
 
 static char *key = NULL;
-static int keysize = 0;
 static char *value = NULL;
-static int valsize = 0;
+static size_t keysize = 0;
+static size_t valsize = 0;
 
 #define ALLOCINCR 1024
 
@@ -392,11 +604,9 @@ getrcskey (fp, keyp, valp)
 {
     char *cur, *max;
     int c;
-    int funky = 0;
-    int white = 1;
 
     /* skip leading whitespace */
-    while (1)
+    do
     {
 	c = getc (fp);
 	if (c == EOF)
@@ -405,25 +615,22 @@ getrcskey (fp, keyp, valp)
 	    *valp = (char *) NULL;
 	    return (-1);
 	}
-	if (!isspace (c))
-	    break;
-    }
+    } while (whitespace (c));
 
     /* fill in key */
     cur = key;
     max = key + keysize;
-    while (!isspace (c) && c != ';')
+    while (!whitespace (c) && c != ';')
     {
-	if (cur < max)
-	    *cur++ = c;
-	else
+	if (cur >= max)
 	{
 	    key = xrealloc (key, keysize + ALLOCINCR);
 	    cur = key + keysize;
 	    keysize += ALLOCINCR;
 	    max = key + keysize;
-	    *cur++ = c;
 	}
+	*cur++ = c;
+
 	c = getc (fp);
 	if (c == EOF)
 	{
@@ -432,6 +639,13 @@ getrcskey (fp, keyp, valp)
 	    return (-1);
 	}
     }
+    if (cur >= max)
+    {
+	key = xrealloc (key, keysize + ALLOCINCR);
+	cur = key + keysize;
+	keysize += ALLOCINCR;
+	max = key + keysize;
+    }
     *cur = '\0';
 
     /* if we got "desc", we are done with the file */
@@ -442,6 +656,18 @@ getrcskey (fp, keyp, valp)
 	return (-1);
     }
 
+    /* skip whitespace between key and val */
+    while (whitespace (c))
+    {
+	c = getc (fp);
+	if (c == EOF)
+	{
+	    *keyp = (char *) NULL;
+	    *valp = (char *) NULL;
+	    return (-1);
+	}
+    } 
+
     /* if we ended key with a semicolon, there is no value */
     if (c == ';')
     {
@@ -451,33 +677,25 @@ getrcskey (fp, keyp, valp)
     }
 
     /* otherwise, there might be a value, so fill it in */
-    (void) ungetc (c, fp);
     cur = value;
     max = value + valsize;
 
     /* process the value */
     for (;;)
     {
-	/* get a character */
-	c = getc (fp);
-	if (c == EOF)
+	/* handle RCS "strings" */
+	if (c == '@') 
 	{
-	    *keyp = (char *) NULL;
-	    *valp = (char *) NULL;
-	    return (-1);
-	}
-
-	/* if we are in funky mode, do the rest of this string */
-	if (funky)
-	{
-
-	    /*
-	     * funky mode processing does the following: o @@ means one @ o
-	     * all other characters are literal up to a single @ (including
-	     * ';')
-	     */
 	    for (;;)
 	    {
+		c = getc (fp);
+		if (c == EOF)
+		{
+		    *keyp = (char *) NULL;
+		    *valp = (char *) NULL;
+		    return (-1);
+		}
+
 		if (c == '@')
 		{
 		    c = getc (fp);
@@ -487,16 +705,11 @@ getrcskey (fp, keyp, valp)
 			*valp = (char *) NULL;
 			return (-1);
 		    }
+		    
 		    if (c != '@')
-		    {
-			/* @ followed by non @ turns off funky mode */
-			funky = 0;
 			break;
-		    }
-		    /* otherwise, we already ate one @ so copy the other one */
 		}
 
-		/* put the character on the value (maybe allocating space) */
 		if (cur >= max)
 		{
 		    value = xrealloc (value, valsize + ALLOCINCR);
@@ -505,6 +718,13 @@ getrcskey (fp, keyp, valp)
 		    max = value + valsize;
 		}
 		*cur++ = c;
+	    }
+	}
+
+	/* compress whitespace down to a single space */
+	if (whitespace (c))
+	{
+	    do {
 		c = getc (fp);
 		if (c == EOF)
 		{
@@ -512,45 +732,8 @@ getrcskey (fp, keyp, valp)
 		    *valp = (char *) NULL;
 		    return (-1);
 		}
-	    }
-	}
+	    } while (whitespace (c));
 
-	/* if we got the semi-colon we are done with the entire value */
-	if (c == ';')
-	    break;
-
-	/* process the character we got */
-	if (white && c == '@')
-	{
-
-	    /*
-	     * if we are starting a word with an '@', enable funky processing
-	     */
-	    white = 0;			/* you can't be funky and white :-) */
-	    funky = 1;
-	}
-	else
-	{
-
-	    /*
-	     * we put the character on the list, compressing all whitespace
-	     * to a single space
-	     */
-
-	    /* whitespace with white set means compress it out */
-	    if (white && isspace (c))
-		continue;
-
-	    if (isspace (c))
-	    {
-		/* make c a space and set white */
-		white = 1;
-		c = ' ';
-	    }
-	    else
-		white = 0;
-
-	    /* put the char on the end of value (maybe allocating space) */
 	    if (cur >= max)
 	    {
 		value = xrealloc (value, valsize + ALLOCINCR);
@@ -558,17 +741,40 @@ getrcskey (fp, keyp, valp)
 		valsize += ALLOCINCR;
 		max = value + valsize;
 	    }
-	    *cur++ = c;
+	    *cur++ = ' ';
+	}
+
+	/* if we got a semi-colon we are done with the entire value */
+	if (c == ';')
+	    break;
+
+	if (cur >= max)
+	{
+	    value = xrealloc (value, valsize + ALLOCINCR);
+	    cur = value + valsize;
+	    valsize += ALLOCINCR;
+	    max = value + valsize;
+	}
+	*cur++ = c;
+
+	c = getc (fp);
+	if (c == EOF)
+	{
+	    *keyp = (char *) NULL;
+	    *valp = (char *) NULL;
+	    return (-1);
 	}
     }
 
-    /* if the last char was white space, take it off */
-    if (white && cur != value)
-	cur--;
-
     /* terminate the string */
-    if (cur)
-	*cur = '\0';
+    if (cur >= max)
+    {
+	value = xrealloc (value, valsize + ALLOCINCR);
+	cur = value + valsize;
+	valsize += ALLOCINCR;
+	max = value + valsize;
+    }
+    *cur = '\0';
 
     /* if the string is empty, make it null */
     if (value && *value != '\0')
@@ -594,7 +800,7 @@ do_symbols (list, val)
     for (;;)
     {
 	/* skip leading whitespace */
-	while (isspace (*cp))
+	while (whitespace (*cp))
 	    cp++;
 
 	/* if we got to the end, we are done */
@@ -603,10 +809,10 @@ do_symbols (list, val)
 
 	/* split it up into tag and rev */
 	tag = cp;
-	cp = index (cp, ':');
+	cp = strchr (cp, ':');
 	*cp++ = '\0';
 	rev = cp;
-	while (!isspace (*cp) && *cp != '\0')
+	while (!whitespace (*cp) && *cp != '\0')
 	    cp++;
 	if (*cp != '\0')
 	    *cp++ = '\0';
@@ -634,7 +840,7 @@ do_branches (list, val)
     for (;;)
     {
 	/* skip leading whitespace */
-	while (isspace (*cp))
+	while (whitespace (*cp))
 	    cp++;
 
 	/* if we got to the end, we are done */
@@ -643,7 +849,7 @@ do_branches (list, val)
 
 	/* find the end of this branch */
 	branch = cp;
-	while (!isspace (*cp) && *cp != '\0')
+	while (!whitespace (*cp) && *cp != '\0')
 	    cp++;
 	if (*cp != '\0')
 	    *cp++ = '\0';
@@ -686,7 +892,7 @@ RCS_getversion (rcs, tag, date, force_tag_match)
 	if (tagrev == NULL)
 	    return ((char *) NULL);
 
-	if ((cp = rindex (tagrev, '.')) != NULL)
+	if ((cp = strrchr (tagrev, '.')) != NULL)
 	    *cp = '\0';
 	rev = RCS_getdatebranch (rcs, date, tagrev);
 	free (tagrev);
@@ -721,11 +927,17 @@ RCS_gettag (rcs, tag, force_tag_match)
     if (rcs == NULL)
 	return ((char *) NULL);
 
+    /* XXX this is probably not necessary, --jtc */
+    if (rcs->flags & PARTIAL) 
+	RCS_reparsercsfile (rcs);
+
     /* If tag is "HEAD", special case to get head RCS revision */
     if (tag && (strcmp (tag, TAG_HEAD) == 0 || *tag == '\0'))
+#if 0 /* This #if 0 is only in the Cygnus code.  Why?  Death support?  */
 	if (force_tag_match && (rcs->flags & VALID) && (rcs->flags & INATTIC))
 	    return ((char *) NULL);	/* head request for removed file */
 	else
+#endif
 	    return (RCS_head (rcs));
 
     if (!isdigit (tag[0]))
@@ -733,8 +945,9 @@ RCS_gettag (rcs, tag, force_tag_match)
 	/* If we got a symbolic tag, resolve it to a numeric */
 	if (rcs == NULL)
 	    p = NULL;
-	else
-	    p = findnode (rcs->symbols, tag);
+	else {
+	    p = findnode (RCS_symbols(rcs), tag);
+	}
 	if (p != NULL)
 	{
 	    int dots;
@@ -750,7 +963,7 @@ RCS_gettag (rcs, tag, force_tag_match)
 	    dots = numdots (tag);
 	    if (dots > 2 && (dots & 1) != 0)
 	    {
-		branch = rindex (tag, '.');
+		branch = strrchr (tag, '.');
 		cp = branch++ - 1;
 		while (*cp != '.')
 		    cp--;
@@ -874,7 +1087,7 @@ RCS_magicrev (rcs, rev)
 	(void) sprintf (xrev, "%s.%d.%d", rev, RCS_MAGIC_BRANCH, rev_num);
 
 	/* walk the symbols list to see if a magic one already exists */
-	if (walklist (rcs->symbols, checkmagic_proc) != 0)
+	if (walklist (RCS_symbols(rcs), checkmagic_proc, NULL) != 0)
 	    continue;
 
 	/* we found a free magic branch.  Claim it as ours */
@@ -887,8 +1100,9 @@ RCS_magicrev (rcs, rev)
  * Returns 0 if the symbol does not match, 1 if it does.
  */
 static int
-checkmagic_proc (p)
+checkmagic_proc (p, closure)
     Node *p;
+    void *closure;
 {
     if (strcmp (check_rev, p->data) == 0)
 	return (1);
@@ -897,9 +1111,9 @@ checkmagic_proc (p)
 }
 
 /*
- * Returns non-zero if the specified revision number or symbolic tag
- * resolves to a "branch" within the rcs file.  We do take into account
- * any magic branches as well.
+ * Given a list of RCSNodes, returns non-zero if the specified
+ * revision number or symbolic tag resolves to a "branch" within the
+ * rcs file.
  */
 int
 RCS_isbranch (file, rev, srcfiles)
@@ -907,7 +1121,6 @@ RCS_isbranch (file, rev, srcfiles)
     char *rev;
     List *srcfiles;
 {
-    int dots;
     Node *p;
     RCSNode *rcs;
 
@@ -922,7 +1135,27 @@ RCS_isbranch (file, rev, srcfiles)
 
     /* now, look for a match in the symbols list */
     rcs = (RCSNode *) p->data;
-    p = findnode (rcs->symbols, rev);
+    return (RCS_nodeisbranch (rev, rcs));
+}
+
+/*
+ * Given an RCSNode, returns non-zero if the specified revision number
+ * or symbolic tag resolves to a "branch" within the rcs file.  We do
+ * take into account any magic branches as well.
+ */
+int
+RCS_nodeisbranch (rev, rcs)
+    char *rev;
+    RCSNode *rcs;
+{
+    int dots;
+    Node *p;
+
+    /* numeric revisions are easy -- even number of dots is a branch */
+    if (isdigit (*rev))
+	return ((numdots (rev) & 1) == 0);
+
+    p = findnode (RCS_symbols(rcs), rev);
     if (p == NULL)
 	return (0);
     dots = numdots (p->data);
@@ -933,7 +1166,7 @@ RCS_isbranch (file, rev, srcfiles)
     if (dots > 2)
     {
 	char *magic;
-	char *branch = rindex (p->data, '.');
+	char *branch = strrchr (p->data, '.');
 	char *cp = branch - 1;
 	while (*cp != '.')
 	    cp--;
@@ -972,7 +1205,7 @@ RCS_whatbranch (file, rev, srcfiles)
 
     /* now, look for a match in the symbols list */
     rcs = (RCSNode *) p->data;
-    p = findnode (rcs->symbols, rev);
+    p = findnode (RCS_symbols(rcs), rev);
     if (p == NULL)
 	return ((char *) NULL);
     dots = numdots (p->data);
@@ -983,7 +1216,7 @@ RCS_whatbranch (file, rev, srcfiles)
     if (dots > 2)
     {
 	char *magic;
-	char *branch = rindex (p->data, '.');
+	char *branch = strrchr (p->data, '.');
 	char *cp = branch++ - 1;
 	while (*cp != '.')
 	    cp--;
@@ -1008,7 +1241,7 @@ RCS_whatbranch (file, rev, srcfiles)
  * Get the head of the specified branch.  If the branch does not exist,
  * return NULL or RCS_head depending on force_tag_match
  */
-static char *
+char *
 RCS_getbranch (rcs, tag, force_tag_match)
     RCSNode *rcs;
     char *tag;
@@ -1024,8 +1257,11 @@ RCS_getbranch (rcs, tag, force_tag_match)
     if (rcs == NULL)
 	return ((char *) NULL);
 
+    if (rcs->flags & PARTIAL)
+	RCS_reparsercsfile (rcs);
+
     /* find out if the tag contains a dot, or is on the trunk */
-    cp = rindex (tag, '.');
+    cp = strrchr (tag, '.');
 
     /* trunk processing is the special case */
     if (cp == NULL)
@@ -1163,6 +1399,9 @@ RCS_getdate (rcs, date, force_tag_match)
     if (rcs == NULL)
 	return ((char *) NULL);
 
+    if (rcs->flags & PARTIAL)
+	RCS_reparsercsfile (rcs);
+
     /* if the head is on a branch, try the branch first */
     if (rcs->branch != NULL)
 	retval = RCS_getdatebranch (rcs, date, rcs->branch);
@@ -1237,13 +1476,19 @@ RCS_getdatebranch (rcs, date, branch)
 
     /* look up the first revision on the branch */
     xrev = xstrdup (branch);
-    cp = rindex (xrev, '.');
+    cp = strrchr (xrev, '.');
     if (cp == NULL)
     {
 	free (xrev);
 	return (NULL);
     }
     *cp = '\0';				/* turn it into a revision */
+
+    assert (rcs != NULL);
+
+    if (rcs->flags & PARTIAL)
+	RCS_reparsercsfile (rcs);
+
     p = findnode (rcs->versions, xrev);
     free (xrev);
     if (p == NULL)
@@ -1327,6 +1572,9 @@ RCS_getrevtime (rcs, rev, date, fudge)
     if (rcs == NULL)
 	return (revdate);
 
+    if (rcs->flags & PARTIAL)
+	RCS_reparsercsfile (rcs);
+
     /* look up the revision */
     p = findnode (rcs->versions, rev);
     if (p == NULL)
@@ -1374,6 +1622,25 @@ RCS_getrevtime (rcs, rev, date, fudge)
     return (revdate);
 }
 
+List *
+RCS_symbols(rcs)
+    RCSNode *rcs;
+{
+    assert(rcs != NULL);
+
+    if (rcs->flags & PARTIAL)
+	RCS_reparsercsfile (rcs);
+
+    if (rcs->symbols_data) {
+	rcs->symbols = getlist ();
+	do_symbols (rcs->symbols, rcs->symbols_data);
+	free(rcs->symbols_data);
+	rcs->symbols_data = NULL;
+    }
+
+    return rcs->symbols;
+}
+
 /*
  * The argument ARG is the getopt remainder of the -k option specified on the
  * command line.  This function returns malloc'ed space that can be used
@@ -1381,12 +1648,24 @@ RCS_getrevtime (rcs, rev, date, fudge)
  */
 char *
 RCS_check_kflag (arg)
-    char *arg;
+    const char *arg;
 {
-    static char *kflags[] =
-    {"kv", "kvl", "k", "v", "o", (char *) NULL};
+    static const char *const kflags[] =
+    {"kv", "kvl", "k", "v", "o", "b", (char *) NULL};
+    static const char *const  keyword_usage[] =
+    {
+      "%s %s: invalid RCS keyword expansion mode\n",
+      "Valid expansion modes include:\n",
+      "   -kkv\tGenerate keywords using the default form.\n",
+      "   -kkvl\tLike -kkv, except locker's name inserted.\n",
+      "   -kk\tGenerate only keyword names in keyword strings.\n",
+      "   -kv\tGenerate only keyword values in keyword strings.\n",
+      "   -ko\tGenerate the old keyword string (no changes from checked in file).\n",
+      "   -kb\tGenerate binary file unmodified (merges not allowed) (RCS 5.7).\n",
+      NULL,
+    };
     char karg[10];
-    char **cpp = NULL;
+    char const *const *cpp = NULL;
 
 #ifndef HAVE_RCS5
     error (1, 0, "%s %s: your version of RCS does not support the -k option",
@@ -1404,12 +1683,7 @@ RCS_check_kflag (arg)
 
     if (arg == NULL || *cpp == NULL)
     {
-	(void) fprintf (stderr, "%s %s: invalid -k option\n",
-			program_name, command_name);
-	(void) fprintf (stderr, "\tvalid options are:\n");
-	for (cpp = kflags; *cpp != NULL; cpp++)
-	    (void) fprintf (stderr, "\t\t-k%s\n", *cpp);
-	error (1, 0, "Please retry with a valid -k option");
+	usage (keyword_usage);
     }
 
     (void) sprintf (karg, "-k%s", *cpp);
@@ -1422,10 +1696,10 @@ RCS_check_kflag (arg)
  */
 void
 RCS_check_tag (tag)
-    char *tag;
+    const char *tag;
 {
     char *invalid = "$,.:;@";		/* invalid RCS tag characters */
-    char *cp;
+    const char *cp;
 
     /*
      * The first character must be an alphabetic letter. The remaining
@@ -1439,7 +1713,7 @@ RCS_check_tag (tag)
 	    if (!isgraph (*cp))
 		error (1, 0, "tag `%s' has non-visible graphic characters",
 		       tag);
-	    if (index (invalid, *cp))
+	    if (strchr (invalid, *cp))
 		error (1, 0, "tag `%s' must not contain the characters `%s'",
 		       tag, invalid);
 	}
@@ -1447,3 +1721,27 @@ RCS_check_tag (tag)
     else
 	error (1, 0, "tag `%s' must start with a letter", tag);
 }
+
+#ifdef DEATH_SUPPORT
+/*
+ * Return true if RCS revision with TAG is a dead revision.
+ */
+int
+RCS_isdead (rcs, tag)
+    RCSNode *rcs;
+    const char *tag;
+{
+    Node *p;
+    RCSVers *version;
+
+    if (rcs->flags & PARTIAL)
+	RCS_reparsercsfile (rcs);
+
+    p = findnode (rcs->versions, tag);
+    if (p == NULL)
+	return (0);
+
+    version = (RCSVers *) p->data;
+    return (version->dead);
+}
+#endif /* DEATH_SUPPORT */
